package gregapi.old;

import static gregapi.data.CS.B;
import static gregapi.data.CS.F;
import static gregapi.data.CS.T;
import gregapi.GT_API;
import gregapi.code.ItemStackContainer;
import gregapi.code.ItemStackMap;
import gregapi.code.ItemStackSet;
import gregapi.config.Config;
import gregapi.old.Textures.BlockIcons;
import gregapi.old.interfaces.IDamagableItem;
import gregapi.old.interfaces.metatileentity.IMetaTileEntity;
import gregapi.recipes.GT_ModHandler;
import gregapi.render.BlockTextureDefault;
import gregapi.render.ITexture;
import gregapi.util.UT;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import net.minecraft.block.Block;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.item.ItemStack;
import net.minecraft.world.World;

/**
 * @author Gregorius Techneticies
 * 
 * Please do not include this File in your Mod-download as it ruins compatiblity, like with the IC2-API
 * You may just copy those Functions into your Code, or better call them via reflection.
 * 
 * The whole API is the basic construct of my Mod. Everything is dependent on it.
 * I change things quite often so please don't include any File inside your Mod, even if it is an Interface.
 * Since some Authors were stupid enough to break this simple Rule, I added Version checks to enforce it.
 * 
 * In these Folders are many useful Functions. You can use them via reflection if you want.
 * I know not everything is compilable due to API's of other Mods, but these are easy to fix in your Setup.
 * 
 * You can use this to learn about Modding, but I would recommend simpler Mods.
 * You may even copypaste Code from these API-Files into your Mod, as I have nothing against that, but you should look exactly at what you are copying.
 */
public class GregTech_API {
	/** The Configuration Objects */
	@Deprecated public static Config sRecipeFile = null, sMachineFile = null, sWorldgenFile = null, sMaterialProperties = null, sSpecialFile = null, sClientDataFile = null, sOPStuff = null;
	
	/** My Creative Tab */
	public static final CreativeTabs TAB_GREGTECH = new GT_CreativeTab("Main", "Main");
	
	public static int TICKS_FOR_LAG_AVERAGING = 25, MILLISECOND_THRESHOLD_UNTIL_LAG_WARNING = 100;
	
	/**
	 * A List of all registered MetaTileEntities
	 * 
	 *      0 -  1199 are used by GregTech.
	 *   1200 -  2047 are used for GregTech Cables.
	 *   2048 -  2559 are reserved for OvermindDL.
	 *   2560 -  3071 are reserved for Immibis.
	 *   3072 -  3583 are reserved for LinusPhoenix.
	 *   3584 -  4095 are reserved for BloodyAsp.
	 *   4096 -  5095 are used for GregTech Frames.
	 *   5096 -  6099 are used for GregTech Pipes.
	 *   6100 -  8191 are used for GregTech Decoration Blocks.
	 *   8192 -  8703 are reserved for ZL123.
	 *   8704 -  9215 are reserved for Mr10Movie.
	 *   9216 -  9727 are used for GregTech Automation Machines.
	 *   9728 - 10239 are reserved for 28Smiles.
	 *  10240 - 10751 are reserved for VirMan.
	 *  10752 - 11263 are reserved for the next one who asks me.
	 *  11264 - 32766 are currently free.
	 *  
	 * Contact me if you need a free ID-Range, which doesn't conflict with other Addons.
	 * You could make an ID-Config, but we all know, what "stupid" customers think about conflicting ID's
	 */
	public static final IMetaTileEntity[] METATILEENTITIES = new IMetaTileEntity[Short.MAX_VALUE - 1];
	
	/** Initialized by the Block creation. */
	public static Block sBlockMachines;
	public static Block sBlockOres1;
	public static Block sBlockGranites, sBlockConcretes;
	public static Block sBlockCasings1, sBlockCasings2, sBlockCasings3, sBlockCasings4;
	
	/** Getting assigned by the Config */
	public static boolean sDrinksAlwaysDrinkable = F, sMultiThreadedSounds = F, sColoredGUI = T, sConstantEnergy = T, sMachineExplosions = T, sMachineFlammable = T, sMachineNonWrenchExplosions = T, sMachineRainExplosions = T, sMachineThunderExplosions = T, sMachineFireExplosions = T, sMachineWireFire = true;
	
	/** Getting assigned by the Mod loading */
	public static boolean sUnificationEntriesRegistered = F;
	
	/** The Icon List for Covers */
	public static final Map<ItemStackContainer, ITexture> sCovers = new ItemStackMap<ItemStackContainer, ITexture>();
	
	/** The List of Cover Behaviors for the Covers */
	public static final Map<ItemStackContainer, GT_CoverBehavior> sCoverBehaviors = new ItemStackMap<ItemStackContainer, GT_CoverBehavior>();
	
	/** The List of Circuit Behaviors for the Redstone Circuit Block */
	public static final Map<Integer, GT_CircuitryBehavior> sCircuitryBehaviors = new HashMap<Integer, GT_CircuitryBehavior>();
	
	/** The List of Blocks, which can conduct Machine Block Updates */
    public static final Map<Block, Integer> sMachineIDs = new HashMap<Block, Integer>();
    
	/** The Redstone Frequencies */
    public static final Map<Integer, Byte> sWirelessRedstone = new HashMap<Integer, Byte>();
    
	/** The IDSU Frequencies */
	public static final Map<Integer, Integer> sIDSUList = new HashMap<Integer, Integer>();
	
	/** The List of all Sounds used in GT, indices are in the static Block at the bottom */
	public static final Map<Integer, String> sSoundList = new HashMap<Integer, String>();
	
	/** The List of Tools, which can be used. Accepts regular damageable Items and Electric Items */
	public static final ItemStackSet<ItemStackContainer> sToolList = new ItemStackSet<ItemStackContainer>(), sCrowbarList = new ItemStackSet<ItemStackContainer>(), sScrewdriverList = new ItemStackSet<ItemStackContainer>(), sWrenchList = new ItemStackSet<ItemStackContainer>(), sSoftHammerList = new ItemStackSet<ItemStackContainer>(), sHardHammerList = new ItemStackSet<ItemStackContainer>(), sSolderingToolList = new ItemStackSet<ItemStackContainer>(), sSolderingMetalList = new ItemStackSet<ItemStackContainer>();
	/** The List of Hazmat Armors */
	public static final ItemStackSet<ItemStackContainer> sGasHazmatList = new ItemStackSet<ItemStackContainer>(), sBioHazmatList = new ItemStackSet<ItemStackContainer>(), sFrostHazmatList = new ItemStackSet<ItemStackContainer>(), sHeatHazmatList = new ItemStackSet<ItemStackContainer>(), sRadioHazmatList = new ItemStackSet<ItemStackContainer>(), sElectroHazmatList = new ItemStackSet<ItemStackContainer>();
	
	/** 
	 * The List of Dimensions, which are Whitelisted for the Teleporter. This list should not contain other Planets.
	 * Mystcraft Dimensions and other Dimensional Things should be allowed.
	 * Mystcraft and Twilight Forest are automatically considered a Dimension, without being in this List.
	 */
	public static final Collection<Integer> sDimensionalList = new HashSet<Integer>();
	
	/**
	 * Causes a Machineblock Update
	 * This update will cause surrounding MultiBlock Machines to update their Configuration.
	 * You should call this Function in @Block.breakBlock and in @Block.onBlockAdded of your Machine.
	 * @param aWorld is being the World
	 * @param aX is the X-Coord of the update causing Block
	 * @param aY is the Y-Coord of the update causing Block
	 * @param aZ is the Z-Coord of the update causing Block
	 */
	public static boolean causeMachineUpdate(World aWorld, int aX, int aY, int aZ) {
		if (!aWorld.isRemote) new Thread(new GT_Runnable_MachineBlockUpdate(aWorld, aX, aY, aZ), "Machine Block Updating").start();
		return T;
	}
	
	/**
	 * Adds a Multi-Machine Block, like my Machine Casings for example.
	 * You should call @causeMachineUpdate in @Block.breakBlock and in @Block.onBlockAdded of your registered Block.
	 * You don't need to register TileEntities which implement @IMachineBlockUpdateable
	 * @param aID the ID of your Block
	 * @param aMeta the Metadata of the Blocks as Bitmask! -1 or ~0 for all Metavalues
	 */
	public static boolean registerMachineBlock(Block aBlock, int aMeta) {
		if (aBlock == null) return F;
        if (GT_API.sCompatTC != null) GT_API.sCompatTC.registerPortholeBlacklistedBlock(aBlock);
		sMachineIDs.put(aBlock, aMeta);
		return T;
	}
	
	/**
	 * Like above but with boolean Parameters instead of a BitMask
	 */
	public static boolean registerMachineBlock(Block aBlock, boolean... aMeta) {
		if (aBlock == null || aMeta == null || aMeta.length == 0) return F;
        if (GT_API.sCompatTC != null) GT_API.sCompatTC.registerPortholeBlacklistedBlock(aBlock);
		int rMeta = 0;
		for (byte i = 0; i < aMeta.length && i < 16; i++) if (aMeta[i]) rMeta |= B[i];
		sMachineIDs.put(aBlock, rMeta);
		return T;
	}
	
	/**
	 * if this Block is a Machine Update Conducting Block
	 */
	public static boolean isMachineBlock(Block aBlock, int aMeta) {
		if (aBlock == null) return F;
		return (sMachineIDs.containsKey(aBlock) && (sMachineIDs.get(aBlock) & B[aMeta]) != 0);
	}
	
	public static void registerCover(ItemStack aStack, ITexture aCover, GT_CoverBehavior aBehavior) {
		if (!sCovers.containsKey(new ItemStackContainer(aStack))) sCovers.put(new ItemStackContainer(aStack), aCover==null||!aCover.isValidTexture()?new BlockTextureDefault(BlockIcons.RENDERING_ERROR):aCover);
		if (aBehavior != null) sCoverBehaviors.put(new ItemStackContainer(aStack), aBehavior);
	}
	
	public static void registerCoverBehavior(ItemStack aStack, GT_CoverBehavior aBehavior) {
		sCoverBehaviors.put(new ItemStackContainer(aStack), aBehavior==null?sDefaultBehavior:aBehavior);
	}
	
	/**
	 * Registers multiple Cover Items. I use that for the OreDict Functionality.
	 * @param aBehavior can be null
	 */
	public static void registerCover(Collection<ItemStack> aStackList, ITexture aCover, GT_CoverBehavior aBehavior) {
		if (aCover.isValidTexture()) for (ItemStack tStack : aStackList) registerCover(tStack, aCover, aBehavior);
	}
	
	/**
	 * This is the generic Cover behavior. Used for the default Covers, which have no Behavior.
	 */
	public static final GT_CoverBehavior sDefaultBehavior = new GT_Cover_Default(), sNoBehavior = new GT_Cover_None();
	
	/**
	 * returns a Cover behavior, guaranteed to not return null after preload
	 */
	public static GT_CoverBehavior getCoverBehavior(ItemStack aStack) {
		if (aStack == null || aStack.getItem() == null) return sNoBehavior;
		GT_CoverBehavior rCover = sCoverBehaviors.get(new ItemStackContainer(aStack));
		if (rCover == null) return sDefaultBehavior;
		return rCover;
	}
	
	/**
	 * returns a Cover behavior, guaranteed to not return null
	 */
	public static GT_CoverBehavior getCoverBehavior(int aStack) {
		if (aStack == 0) return sNoBehavior;
		return getCoverBehavior(UT.Stacks.toStack(aStack));
	}
	
	/**
	 * Register a Wrench to be usable on GregTech Machines.
	 * The Wrench MUST have some kind of Durability unlike certain Buildcraft Wrenches.
	 * 
	 * You need to register Tools in the Load Phase, because otherwise the Autodetection will assign a Tool Type in certain Cases during postload (When IToolWrench or similar Interfaces are implemented).
	 * 
	 * -----
	 * 
	 * Returning true at isDamagable was a great Idea, KingLemming. Well played.
	 * Since the OmniWrench is just a Single-Item-Mod, people can choose if they want your infinite durability or not. So that's not really a Problem.
	 * I even have a new Config to autodisable most infinite BC Wrenches (but that one is turned off).
	 * 
	 * One last Bug for you to fix:
	 * My Autoregistration detects Railcrafts Crowbars, Buildcrafts Wrenches and alike, due to their Interfaces.
	 * Guess what now became a Crowbar by accident. Try registering the Wrench at the load phase to prevent things like that from happening.
	 * Yes, I know that "You need to register Tools in the Load Phase"-Part wasn't there before this. Sorry about that.
	 */
	public static boolean registerWrench(ItemStack aTool) {
		return registerTool(aTool, sWrenchList);
	}
	
	/**
	 * Register a Crowbar to extract Covers from Machines
	 * Crowbars are NOT Wrenches btw.
	 * 
	 * You need to register Tools in the Load Phase, because otherwise the Autodetection will assign a Tool Type in certain Cases during postload (When IToolWrench or similar Interfaces are implemented).
	 */
	public static boolean registerCrowbar(ItemStack aTool) {
		return registerTool(aTool, sCrowbarList);
	}
	
	/**
	 * Register a Screwdriver to interact directly with Machines and Covers
	 * Did I mention, that it is intentionally not possible to make a Multitool, which doesn't switch ItemID (like a Mode) all the time?
	 * 
	 * You need to register Tools in the Load Phase, because otherwise the Autodetection will assign a Tool Type in certain Cases during postload (When IToolWrench or similar Interfaces are implemented).
	 */
	public static boolean registerScrewdriver(ItemStack aTool) {
		return registerTool(aTool, sScrewdriverList);
	}
	
	/**
	 * Register a Soft Hammer to interact with Machines
	 * 
	 * You need to register Tools in the Load Phase, because otherwise the Autodetection will assign a Tool Type in certain Cases during postload (When IToolWrench or similar Interfaces are implemented).
	 */
	public static boolean registerSoftHammer(ItemStack aTool) {
		return registerTool(aTool, sSoftHammerList);
	}
	
	/**
	 * Register a Hard Hammer to interact with Machines
	 * 
	 * You need to register Tools in the Load Phase, because otherwise the Autodetection will assign a Tool Type in certain Cases during postload (When IToolWrench or similar Interfaces are implemented).
	 */
	public static boolean registerHardHammer(ItemStack aTool) {
		return registerTool(aTool, sHardHammerList);
	}
	
	/**
	 * Register a Soldering Tool to interact with Machines
	 * 
	 * You need to register Tools in the Load Phase, because otherwise the Autodetection will assign a Tool Type in certain Cases during postload (When IToolWrench or similar Interfaces are implemented).
	 */
	public static boolean registerSolderingTool(ItemStack aTool) {
		return registerTool(aTool, sSolderingToolList);
	}
	
	/**
	 * Register a Soldering Tin to interact with Soldering Tools
	 * 
	 * You need to register Tools in the Load Phase, because otherwise the Autodetection will assign a Tool Type in certain Cases during postload (When IToolWrench or similar Interfaces are implemented).
	 */
	public static boolean registerSolderingMetal(ItemStack aTool) {
		return registerTool(aTool, sSolderingMetalList);
	}
	
	/**
	 * Generic Function to add Tools to the Lists.
	 * Contains all sanity Checks for Tools, like preventing one Tool from being registered for multiple purposes as controls would override each other.
	 */
	public static boolean registerTool(ItemStack aTool, Collection<ItemStackContainer> aToolList) {
		if (aTool == null || UT.Stacks.inList(aTool, sToolList) || (!aTool.getItem().isDamageable() && !GT_ModHandler.isElectricItem(aTool) && !(aTool.getItem() instanceof IDamagableItem))) return F;
		aToolList.add(new ItemStackContainer(UT.Stacks.copyAmount(1, aTool)));
		sToolList.add(new ItemStackContainer(UT.Stacks.copyAmount(1, aTool)));
		return T;
	}
	
	static {
		sDimensionalList.add(-1);
		sDimensionalList.add( 0);
		sDimensionalList.add( 1);
	}
}