package gregapi.recipes;

import static gregapi.data.CS.NBT_RECYCLING_COMPS;
import gregapi.item.MultiItemTool;
import gregapi.old.interfaces.internal.IGT_CraftingRecipe;
import gregapi.util.UT;
import net.minecraft.enchantment.Enchantment;
import net.minecraft.enchantment.EnchantmentHelper;
import net.minecraft.inventory.InventoryCrafting;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.world.World;
import net.minecraftforge.oredict.ShapelessOreRecipe;

public class AdvancedCraftingShapeless extends ShapelessOreRecipe implements IGT_CraftingRecipe {
	public final boolean mDismantleable, mRemovableByGT, mKeepingNBT;
	private final Enchantment[] mEnchantmentsAdded;
	private final int[] mEnchantmentLevelsAdded;
	
	public AdvancedCraftingShapeless(ItemStack aResult, boolean aDismantleAble, boolean aRemovableByGT, boolean aKeepingNBT, Enchantment[] aEnchantmentsAdded, int[] aEnchantmentLevelsAdded, Object... aRecipe) {
		super(aResult, aRecipe);
		mEnchantmentsAdded = aEnchantmentsAdded;
		mEnchantmentLevelsAdded = aEnchantmentLevelsAdded;
		mRemovableByGT = aRemovableByGT;
		mKeepingNBT = aKeepingNBT;
		mDismantleable = aDismantleAble;
	}
	
	@Override
	public boolean matches(InventoryCrafting aGrid, World aWorld) {
		if (mKeepingNBT) {
			ItemStack tStack = null;
			for (int i = 0; i < aGrid.getSizeInventory(); i++) {
				if (aGrid.getStackInSlot(i) != null) {
					if (tStack != null) {
						if ((tStack.hasTagCompound() != aGrid.getStackInSlot(i).hasTagCompound()) || (tStack.hasTagCompound() && !tStack.getTagCompound().equals(aGrid.getStackInSlot(i).getTagCompound()))) return false;
					}
					tStack = aGrid.getStackInSlot(i);
				}
			}
		}
		return super.matches(aGrid, aWorld);
	}
	
	@Override
	public ItemStack getCraftingResult(InventoryCrafting aGrid) {
		ItemStack rStack = super.getCraftingResult(aGrid);
		if (rStack != null) {
			// Update the Stack
			UT.Stacks.update(rStack);
			
			// Keeping NBT
			if (mKeepingNBT) for (int i = 0; i < aGrid.getSizeInventory(); i++) {
				if (aGrid.getStackInSlot(i) != null && aGrid.getStackInSlot(i).hasTagCompound()) {
					rStack.setTagCompound((NBTTagCompound)aGrid.getStackInSlot(i).getTagCompound().copy());
					break;
				}
			}
			
			// Charge Values
			if (GT_ModHandler.isElectricItem(rStack)) {
				GT_ModHandler.dischargeElectricItem(rStack, Integer.MAX_VALUE, Integer.MAX_VALUE, true, false, true);
				int tCharge = 0;
				for (int i = 0; i < aGrid.getSizeInventory(); i++) tCharge += GT_ModHandler.dischargeElectricItem(aGrid.getStackInSlot(i), Integer.MAX_VALUE, Integer.MAX_VALUE, true, true, true);
				if (tCharge > 0) GT_ModHandler.chargeElectricItem(rStack, tCharge, Integer.MAX_VALUE, true, false);
			}
			
			// Saving Ingredients inside the Item.
			if (mDismantleable) {
				NBTTagCompound rNBT = rStack.getTagCompound(), tNBT = new NBTTagCompound();
				if (rNBT == null) rNBT = new NBTTagCompound();
				for (int i = 0; i < 9; i++) {
					ItemStack tStack = aGrid.getStackInSlot(i);
					if (tStack != null && UT.Stacks.container(tStack, true) == null && !(tStack.getItem() instanceof MultiItemTool)) {
						tStack = UT.Stacks.copyAmount(1, tStack);
						GT_ModHandler.dischargeElectricItem(tStack, Integer.MAX_VALUE, Integer.MAX_VALUE, true, false, true);
						tNBT.setTag(""+i, tStack.writeToNBT(new NBTTagCompound()));
					}
				}
				rNBT.setTag(NBT_RECYCLING_COMPS, tNBT);
				rStack.setTagCompound(rNBT);
			}
			
			// Add Enchantments
			for (int i = 0; i < mEnchantmentsAdded.length; i++) UT.NBT.addEnchantment(rStack, mEnchantmentsAdded[i], EnchantmentHelper.getEnchantmentLevel(mEnchantmentsAdded[i].effectId, rStack) + mEnchantmentLevelsAdded[i]);
			
			// Update the Stack again
			UT.Stacks.update(rStack);
		}
		return rStack;
	}
	
	@Override
	public boolean isRemovable() {
		return mRemovableByGT;
	}
}