package gregapi.tileentity.machines;

import static gregapi.data.CS.*;
import gregapi.GT_API;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_AddToolTips;
import gregapi.block.multitileentity.MultiTileEntityRegistry;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.TagData;
import gregapi.data.LH;
import gregapi.data.LH.Chat;
import gregapi.data.TD;
import gregapi.gui.ContainerClient;
import gregapi.gui.ContainerCommon;
import gregapi.gui.Slot_Output;
import gregapi.gui.Slot_Render;
import gregapi.network.INetworkHandler;
import gregapi.old.Textures;
import gregapi.recipes.Recipe;
import gregapi.recipes.Recipe.RecipeMap;
import gregapi.render.BlockTextureDefault;
import gregapi.render.BlockTextureMulti;
import gregapi.render.IIconContainer;
import gregapi.render.ITexture;
import gregapi.tileentity.DelegatorTileEntity;
import gregapi.tileentity.ITileEntityEnergy;
import gregapi.tileentity.TileEntityBase6Facing;
import gregapi.util.UT;

import java.util.Collection;
import java.util.List;

import net.minecraft.block.Block;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.InventoryPlayer;
import net.minecraft.inventory.ICrafting;
import net.minecraft.inventory.IInventory;
import net.minecraft.inventory.ISidedInventory;
import net.minecraft.inventory.Slot;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraftforge.fluids.*;
import cpw.mods.fml.relauncher.Side;
import cpw.mods.fml.relauncher.SideOnly;

/**
 * @author Gregorius Techneticies
 * 
 * This is the Base Class for almost all my Basic Machines. It is almost too simple to use.
 * 
 * In order to create a Basic Machine (Steel Shredder in this example), I use the following NBT Parameters in my MultiTileEntity System:
 * 
 * NBT_MATERIAL						= MT.Steel.mNameInternal
 * NBT_HARDNESS						= 6.0F
 * NBT_RESISTANCE					= 6.0F
 * NBT_COLOR						= UT.Code.getRGBInt(MT.Steel.mRGBaSolid)
 * NBT_INPUT						= 128
 * NBT_GUI							= RES_PATH_GUI+"machines/Shredder"
 * NBT_TEXTURE						= "shredder"
 * NBT_ENERGY_ACCEPTED				= TD.Energy.RU
 * NBT_RECIPEMAP					= RecipeMap.sShredderRecipes.mNameInternal
 * NBT_EFFICIENCY					= 10000
 * NBT_INV_SIDE_INPUT				= SIDE_BITS[SIDE_TOP]
 * NBT_INV_SIDE_AUTO_INPUT			= SIDE_TOP
 * NBT_INV_SIDE_OUTPUT				= SIDE_BITS[SIDE_BOTTOM]
 * NBT_INV_SIDE_AUTO_OUTPUT			= SIDE_BOTTOM
 * NBT_ENERGY_ACCEPTED_SIDES		= SIDE_BITS[SIDE_LEFT]|SIDE_BITS[SIDE_RIGHT]
 */
public class MultiTileEntityBasicMachine extends TileEntityBase6Facing implements ITileEntityEnergy, ISidedInventory, IFluidHandler, IMTE_AddToolTips {
	protected boolean mActive = F, oActive = F, mRunning = F, oRunning = F, mStateNew = F, mStateOld = F;
	protected boolean[] mEnergyInputAllowed, mItemInputAllowed, mItemOutputAllowed, mFluidInputAllowed, mFluidOutputAllowed;
	protected byte mEnergyInputs = SIDE_BITS[SIDE_BACK], mOutputBlocked = 0;
	protected byte mItemInputs  = SIDE_BITS[SIDE_TOP], mItemOutputs  = SIDE_BITS[SIDE_BOTTOM], mItemAutoInput  = SIDE_TOP, mItemAutoOutput  = SIDE_BOTTOM;
	protected byte mFluidInputs = SIDE_BITS[SIDE_TOP], mFluidOutputs = SIDE_BITS[SIDE_BOTTOM], mFluidAutoInput = SIDE_TOP, mFluidAutoOutput = SIDE_BOTTOM;
	protected short mEfficiency = 10000;
	protected long mEnergy = 0, mInputMin = 16, mInput = 32, mInputMax = 64, mMinEnergy = 0;
	protected TagData mEnergyTypeAccepted = TD.Energy.QU;
	protected Recipe mLastRecipe = null;
	protected FluidTank[] mTanks = ZL_FLUIDTANK;
	protected ItemStack[] mOutputItems = ZL_ITEMSTACK;
	protected FluidStack[] mOutputFluids = ZL_FLUIDSTACK;
	protected IIconContainer[] mTexturesMaterial, mTexturesInactive, mTexturesActive, mTexturesRunning;
	
	public String mGUITexture = "";
	public RecipeMap mRecipes = RecipeMap.sFurnaceRecipes;
	public long mProgress = 0, mMaxProgress = 0;
	public boolean mSuccessful = F;
	
    @Override
	public void readFromNBT2(NBTTagCompound aNBT) {
    	super.readFromNBT2(aNBT);
    	mEnergy = aNBT.getLong(NBT_ENERGY);
        if (aNBT.hasKey(NBT_ACTIVE)) mActive = aNBT.getBoolean(NBT_ACTIVE);
        if (aNBT.hasKey(NBT_RUNNING)) mRunning = aNBT.getBoolean(NBT_RUNNING);
        if (aNBT.hasKey(NBT_STATE+".new")) mStateNew = aNBT.getBoolean(NBT_STATE+".new");
        if (aNBT.hasKey(NBT_STATE+".old")) mStateOld = aNBT.getBoolean(NBT_STATE+".old");
        if (aNBT.hasKey(NBT_EFFICIENCY)) mEfficiency = (short)UT.Code.bind(0, 10000, aNBT.getShort(NBT_EFFICIENCY));
        if (aNBT.hasKey(NBT_INPUT)) {mInput = aNBT.getLong(NBT_INPUT); mInputMin = mInput / 2; mInputMax = mInput * 2;}
        if (aNBT.hasKey(NBT_INPUT_MIN)) {mInputMin = aNBT.getLong(NBT_INPUT_MIN);}
        if (aNBT.hasKey(NBT_INPUT_MAX)) {mInputMax = aNBT.getLong(NBT_INPUT_MAX);}
        if (aNBT.hasKey(NBT_MINENERGY)) {mMinEnergy = aNBT.getLong(NBT_MINENERGY);}
        if (aNBT.hasKey(NBT_PROGRESS)) {mProgress = aNBT.getLong(NBT_PROGRESS);}
        if (aNBT.hasKey(NBT_MAXPROGRESS)) {mMaxProgress = aNBT.getLong(NBT_MAXPROGRESS);}
        if (aNBT.hasKey(NBT_INV_SIDE_INPUT)) mItemInputs = aNBT.getByte(NBT_INV_SIDE_INPUT);
        if (aNBT.hasKey(NBT_INV_SIDE_OUTPUT)) mItemOutputs = aNBT.getByte(NBT_INV_SIDE_OUTPUT);
        if (aNBT.hasKey(NBT_INV_SIDE_AUTO_INPUT)) mItemAutoInput = aNBT.getByte(NBT_INV_SIDE_AUTO_INPUT);
        if (aNBT.hasKey(NBT_INV_SIDE_AUTO_OUTPUT)) mItemAutoOutput = aNBT.getByte(NBT_INV_SIDE_AUTO_OUTPUT);
        if (aNBT.hasKey(NBT_TANK_SIDE_INPUT)) mFluidInputs = aNBT.getByte(NBT_TANK_SIDE_INPUT);
        if (aNBT.hasKey(NBT_TANK_SIDE_OUTPUT)) mFluidOutputs = aNBT.getByte(NBT_TANK_SIDE_OUTPUT);
        if (aNBT.hasKey(NBT_TANK_SIDE_AUTO_INPUT)) mFluidAutoInput = aNBT.getByte(NBT_TANK_SIDE_AUTO_INPUT);
        if (aNBT.hasKey(NBT_TANK_SIDE_AUTO_OUTPUT)) mFluidAutoOutput = aNBT.getByte(NBT_TANK_SIDE_AUTO_OUTPUT);
        if (aNBT.hasKey(NBT_ENERGY_ACCEPTED)) mEnergyTypeAccepted = TagData.createTagData(aNBT.getString(NBT_ENERGY_ACCEPTED));
        if (aNBT.hasKey(NBT_ENERGY_ACCEPTED_SIDES)) mEnergyInputs = aNBT.getByte(NBT_ENERGY_ACCEPTED_SIDES);
        assignSideArrays();
        
        int tCapacity = Math.max(1000, mRecipes.mMaxFluidInputSize * 2);
        if (aNBT.hasKey(NBT_TANK_CAPACITY)) tCapacity = UT.Code.bindInt(aNBT.getLong(NBT_TANK_CAPACITY));
        mTanks = new FluidTank[mRecipes.mInputFluidCount + mRecipes.mOutputFluidCount];
        for (int i = 0; i < mTanks.length; i++) if (aNBT.hasKey(NBT_TANK+"."+i)) mTanks[i] = new FluidTank(tCapacity).readFromNBT(aNBT.getCompoundTag(NBT_TANK+"."+i));
        
        mOutputFluids = new FluidStack[mRecipes.mOutputFluidCount];
		for (int i = 0; i < mOutputFluids.length; i++) mOutputFluids[i] = FluidStack.loadFluidStackFromNBT(aNBT.getCompoundTag(NBT_TANK_OUTPUT+"."+i));
		
        mOutputItems = new ItemStack[mRecipes.mOutputItemsCount];
		for (int i = 0; i < mOutputItems.length; i++) mOutputItems[i] = UT.Stacks.load(aNBT, NBT_INV_OUTPUT+"."+i);
		
		if (GT_API.api_proxy.isClientSide()) {
	        if (aNBT.hasKey(NBT_GUI)) {
	        	mGUITexture = aNBT.getString(NBT_GUI);
		        if (!mGUITexture.endsWith(".png")) mGUITexture += ".png";
	        }
			if (GT_API.sBlockIcons == null && aNBT.hasKey(NBT_TEXTURE)) {
				String tTextureName = aNBT.getString(NBT_TEXTURE);
				mTexturesMaterial = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/colored/back")};
				mTexturesInactive = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay/back")};
				mTexturesActive = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_active/back")};
				mTexturesRunning = new IIconContainer[] {
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/bottom"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/top"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/left"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/front"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/right"),
				new Textures.BlockIcons.CustomIcon("machines/basicmachines/"+tTextureName+"/overlay_running/back")};
			} else {
		        if (getMultiTileEntityRegistryID() != W && getMultiTileEntityID() != W) {
					MultiTileEntityBasicMachine tCanonicalTileEntity = (MultiTileEntityBasicMachine)MultiTileEntityRegistry.getRegistry(getMultiTileEntityRegistryID()).getClassContainer(getMultiTileEntityID()).mCanonicalTileEntity;
					mTexturesMaterial = tCanonicalTileEntity.mTexturesMaterial;
					mTexturesInactive = tCanonicalTileEntity.mTexturesInactive;
					mTexturesRunning  = tCanonicalTileEntity.mTexturesRunning;
					mTexturesActive   = tCanonicalTileEntity.mTexturesActive;
		        } else {
		        	mTexturesMaterial = mTexturesInactive = mTexturesRunning = mTexturesActive = new IIconContainer[6];
		        }
	        }
		}
	}
    
    @Override
	public void writeToNBT2(NBTTagCompound aNBT) {
    	super.writeToNBT2(aNBT);
    	UT.NBT.setNumber(aNBT, NBT_ENERGY, mEnergy);
    	UT.NBT.setNumber(aNBT, NBT_MINENERGY, mMinEnergy);
    	UT.NBT.setNumber(aNBT, NBT_PROGRESS, mProgress);
    	UT.NBT.setNumber(aNBT, NBT_MAXPROGRESS, mMaxProgress);
    	UT.NBT.setBoolean(aNBT, NBT_ACTIVE, mActive);
    	UT.NBT.setBoolean(aNBT, NBT_RUNNING, mRunning);
    	UT.NBT.setBoolean(aNBT, NBT_STATE+".new", mStateNew);
    	UT.NBT.setBoolean(aNBT, NBT_STATE+".old", mStateOld);
    	
        for (int i = 0; i < mTanks.length; i++) aNBT.setTag(NBT_TANK+"."+i, mTanks[i].writeToNBT(new NBTTagCompound()));
    	for (int i = 0; i < mOutputFluids.length; i++) if (mOutputFluids[i] != null) aNBT.setTag(NBT_TANK_OUTPUT+"."+i, mOutputFluids[i].writeToNBT(new NBTTagCompound()));
    	for (int i = 0; i < mOutputItems.length; i++) UT.Stacks.save(aNBT, NBT_INV_OUTPUT+"."+i, mOutputItems[i]);
    }
    
	@Override
	public void addToolTips(List aList, ItemStack aStack, boolean aF3_H) {
		aList.add(Chat.CYAN + LH.get(LH.RECIPES) + ": " + Chat.WHITE + LH.get(mRecipes.mNameInternal));
		aList.add(LH.getToolTipEfficiency(mEfficiency));
		LH.addEnergyToolTips(this, aList, mEnergyTypeAccepted, null, null, null);
		
		// TODO: Input/Output Sides (all 5 types)
	}
	
	@Override
	public long onToolClick2(String aTool, long aRemainingDurability, long aQuality, Entity aPlayer, List<String> aChatReturn, IInventory aPlayerInventory, boolean aSneaking, ItemStack aStack, byte aSide, float aHitX, float aHitY, float aHitZ) {
		long rReturn = super.onToolClick2(aTool, aRemainingDurability, aQuality, aPlayer, aChatReturn, aPlayerInventory, aSneaking, aStack, aSide, aHitX, aHitY, aHitZ);
		if (rReturn > 0) return rReturn;
		
		if (worldObj.isRemote) return 0;
		
		if (aTool.equals(TOOL_softhammer)) {
			mProgress = mMinEnergy = mMaxProgress = 0;
			mOutputFluids = ZL_FLUIDSTACK;
			mOutputItems = ZL_ITEMSTACK;
			return 10000;
		}
		return 0;
	}
	
    @Override
    public void onTickFailed(long aTimer, boolean aIsServerSide) {
    	// Just to prevent Infinite Item dupes in case this happens during the Processing Functions.
		mProgress = mMinEnergy = mMaxProgress = 0;
		mOutputFluids = ZL_FLUIDSTACK;
		mOutputItems = ZL_ITEMSTACK;
    }
    
    @Override
    public void onTick(long aTimer, boolean aIsServerSide) {
    	super.onTick(aTimer, aIsServerSide);
    	
    	if (aIsServerSide) {
    		doWork(aTimer);
    		for (int i = 0; i < mRecipes.mInputFluidCount ; i++) slot(mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1 + i								, UT.Fluids.display(mTanks[i].getFluid(), T));
    		for (int i = 0; i < mRecipes.mOutputFluidCount; i++) slot(mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1 + i + mRecipes.mInputFluidCount	, UT.Fluids.display(mTanks[i + mRecipes.mInputFluidCount].getFluid(), T));
    	}
    }
    
    @Override
    public boolean onTickCheck(long aTimer) {
    	return mActive != oActive || mRunning != oRunning || super.onTickCheck(aTimer);
    }
    
	@Override
    public void onTickResetChecks(long aTimer, boolean aIsServerSide) {
    	super.onTickResetChecks(aTimer, aIsServerSide);
		oRunning = mRunning;
		oActive = mActive;
    }
    
	@Override
	public boolean receiveDataByte(byte aData, INetworkHandler aNetworkHandler) {
		super.receiveDataByte(aData, aNetworkHandler);
		mRunning = ((aData & 2) != 0);
		mActive = ((aData & 1) != 0);
		return T;
	}
	
	@Override
	public boolean onBlockActivated(EntityPlayer aPlayer, byte aSide, float aHitX, float aHitY, float aHitZ) {
		if (super.onBlockActivated(aPlayer, aSide, aHitX, aHitY, aHitZ)) return T;
		if (isServerSide()) openGUI(aPlayer, aSide);
		return T;
	}
	
	@Override
	public long doInject(TagData aEnergyType, byte aSide, long aSize, long aAmount, boolean aDoInject) {
		mStateNew = (aSize > 0);
		aSize = Math.abs(aSize);
		if (aSize > getEnergySizeInputMax(aEnergyType, aSide)) {
			if (aDoInject) overcharge(aSize, aEnergyType);
			return aAmount;
		}
		long tInput = Math.min(mInputMax - mEnergy, aSize * aAmount), tConsumed = Math.min(aAmount, (tInput/aSize) + (tInput%aSize!=0?1:0));
		if (aDoInject) mEnergy += tConsumed * aSize;
		return tConsumed;
	}
	
	@Override public boolean isEnergyType					(TagData aEnergyType, byte aSide, boolean aEmitting) {return !aEmitting && aEnergyType == mEnergyTypeAccepted;}
	@Override public boolean isEnergyAcceptingFrom			(TagData aEnergyType, byte aSide, boolean aTheoretical) {return mEnergyInputAllowed[aSide] && super.isEnergyAcceptingFrom(aEnergyType, aSide, aTheoretical);}
	@Override public long getEnergySizeInputMin				(TagData aEnergyType, byte aSide) {return mInputMin;}
	@Override public long getEnergySizeInputRecommended		(TagData aEnergyType, byte aSide) {return mInput;}
	@Override public long getEnergySizeInputMax				(TagData aEnergyType, byte aSide) {return mInputMax;}
	@Override public Collection<TagData> getEnergyTypes(byte aSide) {return new ArrayListNoNulls<TagData>(F, mEnergyTypeAccepted);}
	
	// Inventory Stuff
	
	@Override
	public ItemStack[] getDefaultInventory(NBTTagCompound aNBT) {
        if (aNBT.hasKey(NBT_RECIPEMAP)) mRecipes = RecipeMap.RECIPE_MAPS.get(aNBT.getString(NBT_RECIPEMAP));
        ACCESSIBLE_SLOTS = new int[mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount];
        for (int i = 0; i < ACCESSIBLE_SLOTS.length; i++) ACCESSIBLE_SLOTS[i] = i;
		return new ItemStack[mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1 + mRecipes.mInputFluidCount + mRecipes.mOutputFluidCount];
	}
	
	protected void assignSideArrays() {
        mItemInputAllowed   = new boolean[] {FACE_CONNECTED[FACING_ROTATIONS[mFacing][0]][mItemInputs  ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][1]][mItemInputs  ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][2]][mItemInputs  ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][3]][mItemInputs  ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][4]][mItemInputs  ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][5]][mItemInputs  ], T};
        mItemOutputAllowed  = new boolean[] {FACE_CONNECTED[FACING_ROTATIONS[mFacing][0]][mItemOutputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][1]][mItemOutputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][2]][mItemOutputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][3]][mItemOutputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][4]][mItemOutputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][5]][mItemOutputs ], T};
        mFluidInputAllowed  = new boolean[] {FACE_CONNECTED[FACING_ROTATIONS[mFacing][0]][mFluidInputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][1]][mFluidInputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][2]][mFluidInputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][3]][mFluidInputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][4]][mFluidInputs ], FACE_CONNECTED[FACING_ROTATIONS[mFacing][5]][mFluidInputs ], T};
        mFluidOutputAllowed = new boolean[] {FACE_CONNECTED[FACING_ROTATIONS[mFacing][0]][mFluidOutputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][1]][mFluidOutputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][2]][mFluidOutputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][3]][mFluidOutputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][4]][mFluidOutputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][5]][mFluidOutputs], T};
        mEnergyInputAllowed = new boolean[] {FACE_CONNECTED[FACING_ROTATIONS[mFacing][0]][mEnergyInputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][1]][mEnergyInputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][2]][mEnergyInputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][3]][mEnergyInputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][4]][mEnergyInputs], FACE_CONNECTED[FACING_ROTATIONS[mFacing][5]][mEnergyInputs], T};
	}
	
	protected int[] ACCESSIBLE_SLOTS;
	@Override public int[] getAccessibleSlotsFromSide(int aSide) {return ACCESSIBLE_SLOTS;}
	@Override public boolean canDrop(int aInventorySlot) {return aInventorySlot < mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount + 1;}
	
	@Override
	public boolean canInsertItem(int aSlot, ItemStack aStack, int aSide) {
		if (!mItemInputAllowed[aSide] || aSlot >= mRecipes.mInputItemsCount) return F;
		for (int i = 0; i < mRecipes.mInputItemsCount; i++) if (UT.Stacks.equal(aStack, slot(i), T)) return i == aSlot;
		return mRecipes.containsInput(aStack);
	}
	
	@Override
	public boolean canExtractItem(int aSlot, ItemStack aStack, int aSide) {
		return mItemOutputAllowed[aSide] && aSlot >= mRecipes.mInputItemsCount && aSlot < mRecipes.mInputItemsCount + mRecipes.mOutputItemsCount;
	}
	
	// Tank things
	
	@Override
	protected IFluidTank getFluidTankFillable(byte aSide, FluidStack aFluidToFill) {
		if (!mFluidInputAllowed[aSide]) return null;
		for (int i = 0; i < mRecipes.mInputFluidCount; i++) if (UT.Fluids.equal(aFluidToFill, mTanks[i].getFluid(), F)) return mTanks[i];
		if (!mRecipes.containsInput(aFluidToFill)) return null;
		for (int i = 0; i < mRecipes.mInputFluidCount; i++) if (mTanks[i].getFluidAmount() == 0) return mTanks[i];
		return null;
	}
	
	@Override
	protected IFluidTank getFluidTankDrainable(byte aSide, FluidStack aFluidToDrain) {
		if (!mFluidOutputAllowed[aSide]) return null;
		if (aFluidToDrain == null) {
			for (int i = 0; i < mRecipes.mOutputFluidCount; i++) if (mTanks[i+mRecipes.mInputFluidCount].getFluidAmount() != 0) return mTanks[i+mRecipes.mInputFluidCount];
		} else {
			for (int i = 0; i < mRecipes.mOutputFluidCount; i++) if (UT.Fluids.equal(aFluidToDrain, mTanks[i+mRecipes.mInputFluidCount].getFluid(), F)) return mTanks[i+mRecipes.mInputFluidCount];
		}
		return null;
	}
	
	@Override
	protected IFluidTank[] getFluidTanks(byte aSide) {
		if (mFluidInputAllowed[aSide]) {
			if (mFluidOutputAllowed[aSide]) return mTanks;
			IFluidTank[] rTanks = new IFluidTank[mRecipes.mInputFluidCount];
			for (int i = 0; i < rTanks.length; i++) rTanks[i] = mTanks[i];
			return rTanks;
		}
		if (mFluidOutputAllowed[aSide]) {
			IFluidTank[] rTanks = new IFluidTank[mRecipes.mOutputFluidCount];
			for (int i = 0; i < rTanks.length; i++) rTanks[i] = mTanks[i+mRecipes.mInputFluidCount];
			return rTanks;
		}
		return ZL_FLUIDTANK;
	}
	
	// Stuff to Override
	
	protected boolean canOutput(Recipe aRecipe) {
    	for (int i = 0, j = mRecipes.mInputItemsCount; i < mRecipes.mOutputItemsCount; i++, j++) if (slot(j) != null) {
        	if (aRecipe.mNeedsEmptyOutput || (aRecipe.mOutputs[i] != null && (!UT.Stacks.equal(slot(j), aRecipe.mOutputs[i], F) || slot(j).stackSize + aRecipe.mOutputs[i].stackSize > slot(j).getMaxStackSize()))) {
            	mOutputBlocked++;
            	return F;
    		}
    	}
    	for (int i = 0, j = mRecipes.mInputFluidCount; i < mRecipes.mOutputFluidCount; i++, j++) if (mTanks[j].getFluidAmount() != 0) {
            if (aRecipe.mNeedsEmptyOutput || (aRecipe.mFluidOutputs[i] != null && (!UT.Fluids.equal(mTanks[j].getFluid(), aRecipe.mFluidOutputs[i], F) || mTanks[j].getFluidAmount() > Math.max(999, aRecipe.mFluidOutputs[i].amount)))) {
               	return F;
        	}
    	}
    	return T;
	}
	
	/** return codes for checkRecipe() */
	protected static final int
	DID_NOT_FIND_RECIPE = 0,
	FOUND_RECIPE_BUT_DID_NOT_MEET_REQUIREMENTS = 1,
	FOUND_AND_SUCCESSFULLY_USED_RECIPE = 2;
	
	/**
	 * Override this to check the Recipes yourself, super calls to this could be useful if you just want to add a special case
	 * I thought about Enum too, but Enum doesn't add support for people adding other return Systems.
	 * Funny how Eclipse marks the word Enum as not correctly spelled.
	 * @return see constants above
	 */
	protected int checkRecipe() {
    	if (mRecipes == null) return DID_NOT_FIND_RECIPE;
    	
    	DelegatorTileEntity<IInventory> tTileEntity = null;
    	
    	byte tAutoInput = FACING_ROTATIONS[mFacing][mItemAutoInput];
		
    	if (SIDES_VALID[tAutoInput]) {
    		tTileEntity = getAdjacentInventory(tAutoInput);
    		if (tTileEntity.mTileEntity == null) tTileEntity = null;
    	}
    	
		ItemStack[] tInputs = new ItemStack[mRecipes.mInputItemsCount];
		for (int i = 0; i < mRecipes.mInputItemsCount; i++) {
			if (tTileEntity != null) UT.Inventories.moveOneItemStackIntoSlot(tTileEntity.mTileEntity, this, tTileEntity.mSideOfTileEntity, i, null, F, (byte)64, (byte)1, (byte)64, (byte)1);
			tInputs[i] = slot(i);
		}
    	
    	Recipe tRecipe = mRecipes.findRecipe(this, mLastRecipe, F, mInputMax, NI, ZL_FLUIDSTACK, tInputs);
    	if (tRecipe == null) return DID_NOT_FIND_RECIPE;
    	if (tRecipe.mCanBeBuffered) mLastRecipe = tRecipe;
        if (!canOutput(tRecipe)) return FOUND_RECIPE_BUT_DID_NOT_MEET_REQUIREMENTS;
        if (!tRecipe.isRecipeInputEqual(true, ZL_FLUIDSTACK, tInputs)) return FOUND_RECIPE_BUT_DID_NOT_MEET_REQUIREMENTS;
        
        mOutputItems = tRecipe.getOutputs(worldObj.rand);
        mOutputFluids = tRecipe.getFluidOutputs(worldObj.rand);
        mMinEnergy = tRecipe.mEUt;
        long tMultiplier = 1;
        while (mMinEnergy < mInputMin && mMinEnergy * 4 <= mInputMax) {mMinEnergy *= 4; tMultiplier *= 2;}
        mMaxProgress = Math.max(1, UT.Code.units(tMultiplier * tRecipe.mEUt * tRecipe.mDuration, mEfficiency, mEnergyTypeAccepted==TD.Energy.RF?RF_PER_EU*10000:mEnergyTypeAccepted==TD.Energy.MJ?(RF_PER_EU*10000)/RF_PER_MJ:10000, T));
        
        removeAllDroppableNullStacks();
        return FOUND_AND_SUCCESSFULLY_USED_RECIPE;
    }
	
	protected void doWork(long aTimer) {
		mSuccessful = F;
		if (mEnergy >= mInputMin && mEnergy >= mMinEnergy) {
			mActive = doActive(aTimer, Math.min(mInputMax, mEnergy));
			mRunning = T;
		} else {
			mActive = doInactive(aTimer);
			mRunning = F;
		}
		mEnergy -= mInputMax; if (mEnergy < 0) mEnergy = 0;
	}
	
	protected boolean doActive(long aTimer, long aEnergy) {
		boolean rActive = F;
		
		if (mMaxProgress > 0) {
			rActive = T;
			if (mProgress <= mMaxProgress) mProgress += aEnergy;
    		if (mProgress >= mMaxProgress && (mStateOld&&!mStateNew || !TD.Energy.ALL_ALTERNATING.contains(mEnergyTypeAccepted))) {
    			for (int i = 0; i < mRecipes.mOutputItemsCount; i++) for (int j = 0; j < mOutputItems .length; j++) if (mOutputItems [j] != null && addStackToSlot(mRecipes.mInputItemsCount+i, mOutputItems[j])) {mOutputItems[j] = null; break;}
    			for (int i = 0; i < mRecipes.mOutputFluidCount; i++) for (int j = 0; j < mOutputFluids.length; j++) if (mOutputFluids[j] != null) {
    				if (mTanks[mRecipes.mInputFluidCount + i].getFluidAmount() == 0) {
    					mTanks[mRecipes.mInputFluidCount + i].setFluid(mOutputFluids[j]);
    					mOutputFluids[j] = null; break;
    				} else if (UT.Fluids.equal(mTanks[mRecipes.mInputFluidCount + i].getFluid(), mOutputFluids[j], F)) {
    					mTanks[mRecipes.mInputFluidCount + i].getFluid().amount += mOutputFluids[j].amount;
    					mOutputFluids[j] = null; break;
    				}
    			}
    			mOutputItems = ZL_ITEMSTACK;
    			mOutputFluids = ZL_FLUIDSTACK;
    			mMinEnergy = 0;
    			mProgress -= mMaxProgress; // this way the leftover energy can be used on the next processed thing.
    			mMaxProgress = 0;
    			mSuccessful = T;
    			onProcessFinished();
    			
    			

    			mProgress = mMinEnergy = mMaxProgress = 0;
    			mOutputItems = ZL_ITEMSTACK;
    			mOutputFluids = ZL_FLUIDSTACK;
    		}
    	}
		
		mStateOld = mStateNew;
		
		boolean
		tRemovedOutputFluid = F;
		
		if (mSuccessful || aTimer%20 == 5) {
			if (SIDES_VALID[mFluidAutoInput]) {
				byte tAutoInput = FACING_ROTATIONS[mFacing][mFluidAutoInput];
				DelegatorTileEntity<IFluidHandler> tTileEntity = getAdjacentTank(tAutoInput);
				if (tTileEntity.mTileEntity != null) {
					FluidTankInfo[] tInfos = tTileEntity.mTileEntity.getTankInfo(FORGE_DIR[tTileEntity.mSideOfTileEntity]);
					if (tInfos != null) for (FluidTankInfo tInfo : tInfos) if (tInfo != null && tInfo.fluid != null && tInfo.fluid.amount > 0 && mRecipes.containsInput(tInfo.fluid)) {
						FluidStack tDrained = tTileEntity.mTileEntity.drain(FORGE_DIR[tTileEntity.mSideOfTileEntity], tInfo.fluid, F);
						if (tDrained != null && fill(FORGE_DIR[mFluidAutoInput], tDrained, F) > 0) fill(FORGE_DIR[mFluidAutoInput], tTileEntity.mTileEntity.drain(FORGE_DIR[tTileEntity.mSideOfTileEntity], tInfo.fluid, T), T);
					}
				}
			}
			if (SIDES_VALID[mFluidAutoOutput]) {
				byte tAutoOutput = FACING_ROTATIONS[mFacing][mFluidAutoOutput];
				DelegatorTileEntity<IFluidHandler> tTileEntity = getAdjacentTank(tAutoOutput);
				if (tTileEntity.mTileEntity != null) for (int i = 0; i < mRecipes.mOutputFluidCount; i++) {
					FluidStack tDrained = mTanks[i+mRecipes.mInputFluidCount].drain(Integer.MAX_VALUE, F);
					if (tDrained != null && tTileEntity.mTileEntity.fill(FORGE_DIR[tTileEntity.mSideOfTileEntity], tDrained, F) > 0 && tTileEntity.mTileEntity.fill(FORGE_DIR[tTileEntity.mSideOfTileEntity], mTanks[i+mRecipes.mInputFluidCount].drain(Integer.MAX_VALUE, T), T) > 0) tRemovedOutputFluid = T;
				}
			}
		}
		
		boolean
		tOutputEmpty = T;
		
		for (int i = mRecipes.mInputItemsCount, j = i + mRecipes.mOutputItemsCount; i < j; i++) if (slot(i) != null) {tOutputEmpty = F; break;}
		
		if (!tOutputEmpty && SIDES_VALID[mItemAutoOutput] && (mSuccessful || mInventoryChanged || !mRunning || mOutputBlocked == 1 || aTimer%1200 == 5)) {
			byte tAutoOutput = FACING_ROTATIONS[mFacing][mItemAutoOutput];
			DelegatorTileEntity<TileEntity> tTileEntity = getAdjacentTileEntity(tAutoOutput);
			for (int i = 0; i < mRecipes.mOutputItemsCount; i++) {
				UT.Inventories.moveOneItemStack(this, tTileEntity, tAutoOutput, tTileEntity.mSideOfTileEntity, null, F, (byte)64, (byte)1, (byte)64, (byte)1);
			}
		}
		
		tOutputEmpty = T;
		
		for (int i = mRecipes.mInputItemsCount, j = i + mRecipes.mOutputItemsCount; i < j; i++) if (slot(i) != null) {tOutputEmpty = F; mOutputBlocked++; break;}
		
		if (tOutputEmpty) mOutputBlocked = 0;
		
    	if (mMaxProgress <= 0) {
    		if ((tRemovedOutputFluid || mSuccessful || mInventoryChanged || !mRunning || aTimer%1200 == 5) && checkRecipe() == FOUND_AND_SUCCESSFULLY_USED_RECIPE) {
		        onProcessStarted();
	    	} else {
	    		mProgress = 0;
	    	}
		}
		
		return rActive;
	}
	
	protected boolean doInactive(long aTimer) {
		mProgress = 0;
		return F;
	}
	
	public class MultiTileEntityGUICommonBasicMachine extends ContainerCommon {
		public MultiTileEntityGUICommonBasicMachine(InventoryPlayer aInventoryPlayer, MultiTileEntityBasicMachine aTileEntity) {
			super(aInventoryPlayer, aTileEntity);
		}
		
	    @Override
		public int addSlots(InventoryPlayer aPlayerInventory) {
	        int tIndex = 0;
	        
	        switch (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputItemsCount) {
	        case  0:
	        	break;
	        case  1:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:25)));
	        	break;
	        case  2:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:25)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:25)));
	        	break;
	        case  3:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:25)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:25)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:25)));
	        	break;
	        case  4:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6?25:34)));
	        	break;
	        case  5:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6?25:34)));
	        	break;
	        case  6:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount>6?25:34)));
	        	break;
	        case  7:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 43));
	        	break;
	        case  8:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 43));
	        	break;
	        case  9:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 43));
	        	break;
	        case 10:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 61));
	        	break;
	        case 11:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 61));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 61));
	        	break;
	        default:
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53,  7));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 25));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 43));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 17, 61));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 35, 61));
	            addSlotToContainer(new Slot(mTileEntity, tIndex++, 53, 61));
	        	break;
	        }
	        
	        switch (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputItemsCount) {
	        case  0:
	        	break;
	        case  1:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:25)));
	        	break;
	        case  2:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:25)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:25)));
	        	break;
	        case  3:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:25)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:25)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:25)));
	        	break;
	        case  4:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6?25:34)));
	        	break;
	        case  5:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6?25:34)));
	        	break;
	        case  6:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6? 7:16)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6?25:34)));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, (((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount>6?25:34)));
	        	break;
	        case  7:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	        	break;
	        case  8:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	        	break;
	        case  9:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	        	break;
	        case 10:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 61));
	        	break;
	        case 11:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 61));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 61));
	        	break;
	        default:
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143,  7));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 25));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 43));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 107, 61));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 125, 61));
	            addSlotToContainer(new Slot_Output(mTileEntity, tIndex++, 143, 61));
	        	break;
	        }
	        
	        for (int i = 0; i < ((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputFluidCount ; i++) addSlotToContainer(new Slot_Render(mTileEntity, tIndex++,  53 - (i%3)*18, 63 - (i/3)*18));
	        for (int i = 0; i < ((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputFluidCount; i++) addSlotToContainer(new Slot_Render(mTileEntity, tIndex++, 107 + (i%3)*18, 63 - (i/3)*18));
	        
	        return super.addSlots(aPlayerInventory);
	    }
	    
	    public short mProgressBar = 0;
	    
	    @Override
		public void detectAndSendChanges() {
	        super.detectAndSendChanges();
	    	for (ICrafting tUpdate : (List<ICrafting>)crafters) {
	    		if (((MultiTileEntityBasicMachine)mTileEntity).mSuccessful) {
	    			tUpdate.sendProgressBarUpdate(this, 0, Short.MAX_VALUE);
	    		} else if (((MultiTileEntityBasicMachine)mTileEntity).mMaxProgress > 0) {
	    			tUpdate.sendProgressBarUpdate(this, 0, (short)UT.Code.units(Math.min(((MultiTileEntityBasicMachine)mTileEntity).mMaxProgress, ((MultiTileEntityBasicMachine)mTileEntity).mProgress), ((MultiTileEntityBasicMachine)mTileEntity).mMaxProgress, Short.MAX_VALUE, T));
	    		} else {
	    			tUpdate.sendProgressBarUpdate(this, 0, -1);
	    		}
	    	}
	    }
	    
	    @Override
		@SideOnly(Side.CLIENT)
	    public void updateProgressBar(int aIndex, int aValue) {
	    	super.updateProgressBar(aIndex, aValue);
	    	switch (aIndex) {
	    	case 0: mProgressBar = (short)aValue; break;
	    	}
	    }
	    
	    @Override public int getSlotStartIndex() {return 0;}
	    @Override public int getShiftClickStartIndex() {return 0;}
	    @Override public int getSlotCount() {return ((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputItemsCount + ((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mOutputItemsCount;}
	    @Override public int getShiftClickSlotCount() {return ((MultiTileEntityBasicMachine)mTileEntity).mRecipes.mInputItemsCount;}
	}
	
	@SideOnly(Side.CLIENT)
	public class MultiTileEntityGUIClientBasicMachine extends ContainerClient {
		public final MultiTileEntityBasicMachine mTileEntity;
		
	    public MultiTileEntityGUIClientBasicMachine(InventoryPlayer aInventoryPlayer, MultiTileEntityBasicMachine aTileEntity) {
	        super(new MultiTileEntityGUICommonBasicMachine(aInventoryPlayer, aTileEntity), aTileEntity.mGUITexture);
	        mTileEntity = aTileEntity;
	        mNEI = mTileEntity.mRecipes.mNameNEI;
	    }
	    
	    @Override
	    protected void drawGuiContainerForegroundLayer(int par1, int par2) {
	    	fontRendererObj.drawString(mTileEntity.hasCustomInventoryName()?mTileEntity.getCustomName():LH.get(mTileEntity.mRecipes.mNameInternal), 8,  4, 4210752);
	    }
	    
	    @Override
	    protected void drawGuiContainerBackgroundLayer(float par1, int par2, int par3) {
	    	super.drawGuiContainerBackgroundLayer(par1, par2, par3);
	        int x = (width - xSize) / 2;
	        int y = (height - ySize) / 2;
	        drawTexturedModalRect(x, y, 0, 0, xSize, ySize);
	        if (mContainer instanceof MultiTileEntityGUICommonBasicMachine) {
	        	if (((MultiTileEntityGUICommonBasicMachine)mContainer).mProgressBar >= 0) {
		        	int tSize = (mTileEntity.mRecipes.mProgressBarDirection % 4 < 2 ? 20 : 18), tProgress = (int)UT.Code.scale(((MultiTileEntityGUICommonBasicMachine)mContainer).mProgressBar, 0, Short.MAX_VALUE, tSize * UT.Code.unsignB(mTileEntity.mRecipes.mProgressBarAmount), F) % (tSize+1);
		        	
		        	switch (mTileEntity.mRecipes.mProgressBarDirection) {
		        	case 0:								drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, tProgress	, 18		); break;
		        	case 1:								drawTexturedModalRect(x + 78 + 20 - tProgress	, y + 24					, 176 + 20 - tProgress	, 0					, tProgress	, 18		); break;
		        	case 2:								drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, 20		, tProgress	); break;
		        	case 3:								drawTexturedModalRect(x + 78					, y + 24 + 18 - tProgress	, 176					, 18 - tProgress	, 20		, tProgress	); break;
		        	case 4: tProgress = 20 - tProgress;	drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, tProgress	, 18		); break;
		        	case 5: tProgress = 20 - tProgress;	drawTexturedModalRect(x + 78 + 20 - tProgress	, y + 24					, 176 + 20 - tProgress	, 0					, tProgress	, 18		); break;
		        	case 6: tProgress = 18 - tProgress;	drawTexturedModalRect(x + 78					, y + 24					, 176					, 0					, 20		, tProgress	); break;
		        	case 7: tProgress = 18 - tProgress;	drawTexturedModalRect(x + 78					, y + 24 + 18 - tProgress	, 176					, 18 - tProgress	, 20		, tProgress	); break;
		        	}
	        	}
	        }
	    }
	}
	
	protected void onProcessStarted() {/**/}
	protected void onProcessFinished() {/**/}
	
	@Override public Object getGUIClient2(int aGUIID, EntityPlayer aPlayer) {return new MultiTileEntityGUIClientBasicMachine(aPlayer.inventory, this);}
	@Override public Object getGUIServer2(int aGUIID, EntityPlayer aPlayer) {return new MultiTileEntityGUICommonBasicMachine(aPlayer.inventory, this);}
	
	@Override public void onFacingChange() {assignSideArrays();}
	@Override public byte getVisualData() {return (byte)((mActive?1:0)|(mRunning?2:0));}
	@Override public byte getDefaultSide() {return SIDE_SOUTH;}
	@Override public boolean[] getValidSides() {return SIDES_HORIZONTAL;}
	@Override public ITexture getTexture2(Block aBlock, int aRenderPass, byte aSide) {return new BlockTextureMulti(new BlockTextureDefault(mTexturesMaterial[FACING_ROTATIONS[mFacing][aSide]], mRGBa), new BlockTextureDefault((mActive||worldObj==null?mTexturesActive:mRunning?mTexturesRunning:mTexturesInactive)[FACING_ROTATIONS[mFacing][aSide]]));}
	
	@Override public String getTileEntityName() {return "gt.multitileentity.machine.basic";}
}